;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: CLIM-CAIRO; -*-
;;; ---------------------------------------------------------------------------
;;;     Title: CAIRO FFI Bindings
;;;   Created: 2005-03-05
;;;    Author: Gilbert Baumann <gilbert@base-engineering.com>
;;;   License: MIT style (see below)
;;; ---------------------------------------------------------------------------
;;;  (c) copyright 2005 by Gilbert Baumann

;;; 
;;;  Permission is hereby granted, free of charge, to any person obtaining
;;;  a copy of this software and associated documentation files (the
;;;  "Software"), to deal in the Software without restriction, including
;;;  without limitation the rights to use, copy, modify, merge, publish,
;;;  distribute, sublicense, and/or sell copies of the Software, and to
;;;  permit persons to whom the Software is furnished to do so, subject to
;;;  the following conditions:
;;; 
;;;  The above copyright notice and this permission notice shall be
;;;  included in all copies or substantial portions of the Software.
;;; 
;;;  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;;;  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;;  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
;;;  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
;;;  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
;;;  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
;;;  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
;;; 

(in-package :CLIM-CAIRO)

#+SBCL
(defmacro def-alien-routine (&rest stuff)
  `(sb-alien:define-alien-routine ,@stuff))

#+SBCL
(defmacro def-alien-type (&rest stuff)
  `(sb-alien:define-alien-type ,@stuff))

(def-alien-type cairo (* (struct cairo)))
(def-alien-type cairo-surface (* (struct cairo-surface)))
(def-alien-type cairo-matrix (* (struct cairo-matrix)))
(def-alien-type cairo-pattern (* (struct pattern)))

;;; Functions for manipulating state objects

(def-alien-routine cairo-create
    cairo
  )

(def-alien-routine cairo-reference
    void
  (cr cairo))

(def-alien-routine cairo-destroy
    void
  (cr cairo))

(def-alien-routine cairo-save
    void
  (cr cairo))

(def-alien-routine cairo-restore
    void
  (cr cairo))

;;; XXX: Replace with cairo_current_gstate/cairo_set_gstate

(def-alien-routine cairo-copy
    void
  (destination cairo) (source cairo))

;;; Modify state

(def-alien-routine cairo-set-target-surface
    void
  (cr cairo) (surface cairo-surface))

(def-alien-type cairo-format
    (enum nil :argb32 :rgb24 :a8 :a1))

(def-alien-routine cairo-set-target-image
    void
  (cr cairo)
  (data (* (unsigned 8)))
  (format cairo-format)
  (width int)
  (height int)
  (stride int))

(def-alien-type cairo-operator
    (enum nil
          :clear
          :src
          :dst
          :over
          :over-reverse
          :in
          :in-reverse
          :out
          :out-reverse
          :atop
          :atop-reverse
          :xor
          :add
          :saturate))

(def-alien-routine cairo-set-operator void
  (cr cairo)
  (op cairo-operator))

;;; Colors

(def-alien-routine ("cairo_set_rgb_color" cairo-set-rgb-color)
    void
  (cr cairo)
  (red double-float)
  (green double-float)
  (blue double-float))

(def-alien-routine ("cairo_set_pattern" cairo-set-pattern)
    void
  (cr cairo)
  (pattern cairo-pattern))

(def-alien-routine ("cairo_set_alpha" cairo-set-alpha)
    void
  (cr cairo)
  (alpha double-float))

(def-alien-routine ("cairo_set_tolerance" cairo-set-tolerance)
    void
  (cr cairo)
  (tolerance double-float))

(def-alien-type cairo-fill-rule
    (enum nil :winding :even-odd))

(def-alien-routine cairo-set-fill-rule
    void
  (cr cairo)
  (fill-rule cairo-fill-rule))

(def-alien-routine cairo-set-line-width
    void
  (cr cairo)
  (w double-float))

(def-alien-type cairo-line-cap
    (enum nil :butt :round :square))

(def-alien-routine cairo-set-line-cap
    void
  (cr cairo) (line-cap cairo-line-cap))

(def-alien-type cairo-line-join
    (enum nil :miter :round :bevel))

(def-alien-routine cairo-set-line-join
    void
  (cr cairo) (line-join cairo-line-join))

(def-alien-routine cairo-set-dash void
  (cr cairo) (dashes (* double-float)) (ndash int) (offset double-float))

(def-alien-routine cairo-set-miter-limit
    int
  (cr cairo) (limit double-float))

;;; Transformations

(def-alien-routine cairo-translate
    void
  (cr cairo) (tx double-float) (ty double-float))

(def-alien-routine cairo-scale
    void
  (cr cairo) (sx double-float) (sy double-float))

(def-alien-routine cairo-rotate
    void
  (cr cairo) (angle double-float))

(def-alien-routine cairo-concat-matrix
    void
  (cr cairo) (matrix cairo-matrix))

(def-alien-routine cairo-set-matrix
    void
  (cr cairo) (matrix cairo-matrix))

(def-alien-routine cairo-default-matrix
    void
  (cr cairo))

(def-alien-routine cairo-identity-matrix
    void
  (cr cairo))

(def-alien-routine cairo-transform-point
    void
  (cr cairo) (x (* double-float)) (y (* double-float)))

(def-alien-routine cairo-transform-distance
    void
  (cr cairo) (dx (* double-float)) (dy (* double-float)))

(def-alien-routine cairo-inverse-transform-point
    void
  (cr cairo) (x (* double-float)) (y (* double-float)))

(def-alien-routine cairo-inverse-transform-distance
    void
  (cr cairo) (dx (* double-float)) (dy (* double-float)))

;;; Path creation functions

(def-alien-routine cairo-new-path
    void
  (cr cairo))

(def-alien-routine cairo-move-to
    void
  (cr cairo) (x double-float) (y double-float))

(def-alien-routine cairo-line-to
    void
  (cr cairo) (x double-float) (y double-float))

(def-alien-routine cairo-curve-to
    void
  (cr cairo)
  (x1 double-float) (y1 double-float)
  (x2 double-float) (y2 double-float)
  (x3 double-float) (y3 double-float))

(def-alien-routine cairo-arc
    void
  (cr cairo)
  (xc double-float) (yc double-float)
  (radius double-float) (angle1 double-float) (angle2 double-float))

(def-alien-routine cairo-arc-negative
    void
  (cr cairo)
  (xc double-float) (yc double-float)
  (radius double-float) (angle1 double-float) (angle2 double-float))

(def-alien-routine cairo-rel-move-to
    void
  (cr cairo) (dx double-float) (dy double-float))

(def-alien-routine cairo-rel-line-to
    void
  (cr cairo) (dx double-float) (dy double-float))

(def-alien-routine cairo-rel-curve-to
    void
  (cr cairo)
  (dx1 double-float) (dy1 double-float)
  (dx2 double-float) (dy2 double-float)
  (dx3 double-float) (dy3 double-float))

(def-alien-routine cairo-rectangle
    void
  (cr cairo) (x double-float) (y double-float) (w double-float) (h double-float))

(def-alien-routine cairo-close-path
    void
  (cr cairo))

(def-alien-routine cairo-stroke
    void
  (cr cairo))

(def-alien-routine cairo-fill
    void
  (cr cairo))

(def-alien-routine cairo-copy-page
    void
  (cr cairo))

(def-alien-routine cairo-show-page
    void
  (cr cairo))

;;; Insideness testing

(def-alien-routine cairo-in-stroke
    int
  (cr cairo) (x double-float) (y double-float))

(def-alien-routine cairo-in-fill
    int
  (cr cairo) (x double-float) (y double-float))

;;; Rectangular extents

(def-alien-routine cairo-stroke-extents
    void
  (cr cairo) (x1 (* double-float)) (y1 (* double-float)) (x2 (* double-float)) (y2 (* double-float)))

(def-alien-routine cairo-fill-extents
    void
  (cr cairo) (x1 (* double-float)) (y1 (* double-float)) (x2 (* double-float)) (y2 (* double-float)))

(def-alien-routine cairo-init-clip
    void
  (cr cairo))

;; Note: cairo_clip does not consume the current path
(def-alien-routine cairo-clip
    void
  (cr cairo))

;;; Font/Text functions

(def-alien-type cairo-font (* (struct cairo-font)))

(def-alien-type cairo-glyph
    (* (struct cairo-glyph
               (index unsigned-int)
               (x double-float)
               (y double-float))))

(def-alien-type cairo-text-extents
    (* (struct cairo-text-extents
               (x-bearing double-float)
               (y-bearing double-float)
               (width double-float)
               (height double-float)
               (x-advance double-float)
               (y-advance double-float))))

(def-alien-type cairo-font-extents
    (* (struct cairo-font-extents
               (ascent double-float)
               (descent double-float)
               (height double-float)
               (max-x-advance double-float)
               (max-y-advance double-float))))

(def-alien-type cairo-font-slant
    (enum nil :normal :italic :oblique))

(def-alien-type cairo-font-weight
    (enum nil :normal :bold))
  
;; This interface is for dealing with text as text, not caring about the
;; font object inside the the cairo_t.

(def-alien-routine cairo-select-font
    void
    (cr cairo)
    (family c-string)
    (slant cairo-font-slant)
    (weight cairo-font-weight))

(def-alien-routine cairo-scale-font
    void
  (cr cairo)
  (scale double-float))

(def-alien-routine cairo-transform-font
    void
  (cr cairo)
  (matrix cairo-matrix))

(def-alien-routine cairo-show-text
    void
  (cr cairo)
  (string c-string))

(def-alien-routine cairo-show-glyphs
    void
  (cr cairo)
  (glyphs cairo-glyph)
  (num-glyphs int))

(def-alien-routine cairo-current-font
    cairo-font
  (cr cairo))

(def-alien-routine cairo-current-font-extents
    void
  (cr cairo)
  (extents cairo-font-extents))

(def-alien-routine cairo-set-font
    void
  (cr cairo)
  (font cairo-font))

(def-alien-routine cairo-text-extents
    void
  (cr cairo)
  (string c-string)              ;### utf-8
  (extents cairo-text-extents))

(def-alien-routine cairo-glyph-extents
    void
  (cr cairo)
  (glyphs cairo-glyph)
  (num-glyphs int)
  (extents cairo-text-extents))

(def-alien-routine cairo-text-path
    void
  (cr cairo)
  (string c-string))             ;### utf-8

(def-alien-routine cairo-glyph-path
    void
  (cr cairo)
  (glyphs cairo-glyph)
  (num-glyphs int))
  
;;; Portable interface to general font features.

(def-alien-routine cairo-font-reference
    void
  (font cairo-font))

(def-alien-routine cairo-font-destroy
    void
  (font cairo-font))

;; ### Yeah, yeah changing APIs ... this cairo-font-set-transform seems to be gone. why?
#+NIL
(def-alien-routine cairo-font-set-transform
    void
  (font cairo-font)
  (matrix cairo-matrix))

#+NIL   ;; ### this too 
(def-alien-routine cairo-font-current-transform
    void
  (font cairo-font)
  (matrix cairo-matrix))

;;; Image functions

(def-alien-routine cairo-show-surface
    void
  (cr cairo)
  (surface cairo-surface)
  (width int)
  (height int))

;;; Query functions 

(def-alien-routine cairo-current-operator
    cairo-operator
  (cr cairo))

(def-alien-routine cairo-current-rgb-color
    void
  (cr cairo)
  (red (* double-float))
  (green (* double-float))
  (blue (* double-float)))

(def-alien-routine cairo-current-pattern
    cairo-pattern
  (cr cairo))

(def-alien-routine cairo-current-alpha
    double-float
  (cr cairo))

(def-alien-routine cairo-current-tolerance
    double-float
  (cr cairo))

(def-alien-routine cairo-current-point
    void
  (cr cairo)
  (x (* double-float))
  (y (* double-float)))

(def-alien-routine cairo-current-fill-rule
    cairo-fill-rule
  (cr cairo))

(def-alien-routine cairo-current-line-width
    double-float
  (cr cairo))

(def-alien-routine cairo-current-line-cap
    cairo-line-cap
  (cr cairo))

(def-alien-routine cairo-current-line-join
    cairo-line-join
  (cr cairo))

(def-alien-routine cairo-current-miter-limit
    double-float
  (cr cairo))

(def-alien-routine cairo-current-matrix
    void
  (cr cairo)
  (matrix cairo-matrix))

(def-alien-routine cairo-current-target-surface
    cairo-surface
  (cr cairo))

#|
typedef void (cairo_move_to_func_t) (void *closure,
				     double x, double y);

typedef void (cairo_line_to_func_t) (void *closure,
				     double x, double y);

typedef void (cairo_curve_to_func_t) (void *closure,
				      double x1, double y1,
				      double x2, double y2,
				      double x3, double y3);

typedef void (cairo_close_path_func_t) (void *closure);

extern void
cairo_current_path (cairo_t			*cr,
		    cairo_move_to_func_t	*move_to,
		    cairo_line_to_func_t	*line_to,
		    cairo_curve_to_func_t	*curve_to,
		    cairo_close_path_func_t	*close_path,
		    void			*closure);

extern void
cairo_current_path_flat (cairo_t			*cr,
			 cairo_move_to_func_t		*move_to,
			 cairo_line_to_func_t		*line_to,
			 cairo_close_path_func_t	*close_path,
			 void				*closure);
|#

;;; Error status queries

(def-alien-type cairo-status
    (enum nil
                :success
                :no-memory
                :invalid-restore
                :invalid-pop-group
                :no-current-font
                :invalid-matrix
                :no-target-surface
                :null-pointer))

(def-alien-routine cairo-status
    cairo-status
  (cr cairo))

(def-alien-routine cairo-status-string
    c-string
  (cr cairo))

;;; Surface manipulation

(def-alien-routine cairo-surface-create-for-image
    cairo-surface
  (data (* (unsigned 8)))
  (format cairo-format)
  (width int)
  (height int)
  (stride int))

(def-alien-routine cairo-surface-create-similar
    cairo-surface
  (other cairo-surface)
  (format cairo-format)
  (width int)
  (height int))

(def-alien-routine cairo-surface-reference
    void
  (surface cairo-surface))

(def-alien-routine cairo-surface-destroy
    void
  (surface cairo-surface))

(def-alien-routine cairo-surface-set-repeat
    cairo-status
  (surface cairo-surface)
  (repeat int))

(def-alien-routine cairo-surface-set-matrix
    cairo-status
  (surface cairo-surface)
  (matrix cairo-matrix))

(def-alien-routine cairo-surface-get-matrix
    cairo-status
  (surface cairo-surface)
  (matrix cairo-matrix))

(def-alien-type cairo-filter
    (enum nil :fast :good :best :nearest :bilinear :gaussian))

(def-alien-routine cairo-surface-set-filter
    cairo-status
  (surface cairo-surface)
  (filter cairo-filter))

(def-alien-routine cairo-surface-get-filter
    cairo-filter
  (surface cairo-surface))

;;; Image-surface functions

(def-alien-routine cairo-image-surface-create
    cairo-surface
  (format cairo-format)
  (width int)
  (height int))

(def-alien-routine cairo-image-surface-create-for-data
    cairo-surface
  (data (* (unsigned 8)))
  (format cairo-format)
  (width int)
  (height int)
  (stride int))

;;; Pattern creation functions

(def-alien-routine cairo-pattern-create-for-surface
    cairo-pattern
  (surface cairo-surface))

(def-alien-routine cairo-pattern-create-linear
    cairo-pattern
  (x1 double-float)
  (y1 double-float)
  (x2 double-float)
  (y2 double-float))

(def-alien-routine cairo-pattern-create-radial
    cairo-pattern
  (cx1 double-float)
  (cy1 double-float)
  (cr1 double-float)
  (cx2 double-float)
  (cy2 double-float)
  (cr2 double-float))

(def-alien-routine cairo-pattern-reference
    void
  (pattern cairo-pattern))

(def-alien-routine cairo-pattern-destroy
    void
  (pattern cairo-pattern))

(def-alien-routine cairo-pattern-add-color-stop
    int                          ;### cairo_status_t
  (pattern cairo-pattern)
  (offset double-float)
  (red double-float)
  (green double-float)
  (blue double-float)
  (alpha double-float))

(def-alien-routine cairo-pattern-set-matrix
    cairo-status
  (pattern cairo-pattern)
  (martix cairo-matrix))

(def-alien-routine cairo-pattern-get-matrix
    cairo-status
  (pattern cairo-pattern)
  (martix cairo-matrix))

(def-alien-type cairo-extend
    (enum nil :none :repeat :reflect))

(def-alien-routine cairo-pattern-set-extend
    cairo-status
  (pattern cairo-pattern)
  (extend cairo-extend))

(def-alien-routine cairo-pattern-get-extend
    cairo-extend
  (pattern cairo-pattern))

(def-alien-routine cairo-pattern-set-filter
    cairo-status
  (pattern cairo-pattern)
  (filter cairo-filter))

(def-alien-routine cairo-pattern-get-filter
    cairo-filter
  (pattern cairo-pattern))

;;; Matrix functions

(def-alien-routine cairo-matrix-create
    cairo-matrix
  )

(def-alien-routine cairo-matrix-destroy
    void
  (matrix cairo-matrix))

(def-alien-routine cairo-matrix-copy
    cairo-status
  (matrix cairo-matrix)
  (other cairo-matrix))

(def-alien-routine cairo-matrix-set-identity
    cairo-status
  (matrix cairo-matrix))

(def-alien-routine cairo-matrix-set-affine
    cairo-status
  (matrix cairo-matrix)
  (a double-float)
  (b double-float)
  (c double-float)
  (d double-float)
  (tx double-float)
  (ty double-float))

(def-alien-routine cairo-matrix-get-affine
    cairo-status
  (matrix cairo-matrix)
  (a (* double-float))
  (b (* double-float))
  (c (* double-float))
  (d (* double-float))
  (tx (* double-float))
  (ty (* double-float)))

(def-alien-routine cairo-matrix-translate
    cairo-status
  (matrix cairo-matrix) (tx double-float) (ty double-float))

(def-alien-routine cairo-matrix-scale
    cairo-status
  (matrix cairo-matrix) (sx double-float) (sy double-float))

(def-alien-routine cairo-matrix-rotate
    cairo-status
  (matrix cairo-matrix) (angle double-float))

(def-alien-routine cairo-matrix-invert
    cairo-status
  (matrix cairo-matrix))

(def-alien-routine cairo-matrix-multiply
    cairo-status
  (result cairo-matrix)
  (a cairo-matrix)
  (b cairo-matrix))

(def-alien-routine cairo-matrix-transform-distance
    cairo-status
  (matrix cairo-matrix) (dx (* double-float)) (dy (* double-float)))

(def-alien-routine cairo-matrix-transform-point
    cairo-status
  (matrix cairo-matrix) (x (* double-float)) (y (* double-float)))

;;;; ------------------------------------------------------------------------------------------

(def-alien-type display (struct display))
(def-alien-type window (unsigned 32))

(def-alien-routine ("cairo_set_target_drawable" cairo-set-target-drawable)
    void
  (cr cairo)
  (display (* display))
  (window window))

;;;; X stuff

(def-alien-routine ("XOpenDisplay" x-open-display)
    (* display)
  (name c-string))

(def-alien-routine ("XCloseDisplay" x-close-display)
    void
  (display (* display)))

(def-alien-routine ("XDefaultRootWindow" x-default-root-window)
    window
  (display (* display)))

(def-alien-routine ("XCreateSimpleWindow" x-create-simple-window)
    window
  (display (* display))
  (parent window)
  (x int) (y int)
  (w int) (h int)
  (bw int)
  (border int)
  (background int))

(def-alien-routine ("XMapWindow" x-map-window)
    int
  (display (* display))
  (window window))

(def-alien-routine ("XFlush" x-flush)
    void
  (display (* display)))

(def-alien-routine ("XSync" x-sync)
    void
  (display (* display))
  (discard int) )

(def-alien-routine ("XSynchronize" x-synchronize)
    int                          ;a function pointer really
  (display (* display))
  (onoff int))

(def-alien-routine cairo-xlib-surface-create
    cairo-surface
  (display (* display))
  (drawable int)
  (visual int)                   ;###
  (format cairo-format)
  (colormap int))                ;###

(def-alien-routine ("XDefaultColormap" x-default-colormap)
    int
  (display (* display))
  (screen int))
